/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.engine.orchestra;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;

import org.jbpm.pvm.env.EnvironmentFactory;
import org.jbpm.pvm.env.Transaction;
import org.ow2.orchestra.deployment.Deployer;
import org.ow2.orchestra.deployment.Deployment;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.parsing.BpelImportParser;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.su.ServiceUnitListener;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.engine.orchestra.util.WSDLAnalyzer;

/**
 * This class implements the orchestra su listener
 *
 * @author Adrien LOUIS - eBMWebSourcing
 * @author Nicolas SALATGE - eBMWebSourcing
 *
 */
public final class OrchestraSUListener implements ServiceUnitListener {

	/**
	 * BPEL property
	 */
	private static final String BPEL_PROPERTY = "bpel";

	/**
	 * JDK logger
	 */
	private final Logger logger;

	/**
	 * JBI Descriptor
	 */
	private Jbi descriptor;

	/**
	 * The environment Factory
	 */
	private final EnvironmentFactory environmentFactory;

	/**
	 * The su root path
	 */
	private String suRootPath;

	private Map<String, BPELProcessEnvironment> bpelProcessMap;

	/**
	 * The default constructor
	 *
	 * @param environmentFactory
	 *            the jbi environnement
	 * @param suManager
	 *            the su manager
	 * @param logger
	 *            the logger
	 */
	public OrchestraSUListener(
			final Map<String, BPELProcessEnvironment> bpelProcessMap,
			final EnvironmentFactory environmentFactory, final Logger logger) {// ,
		// final
		// ClassLoader
		// componentClassLoader)
		// {
		this.logger = logger;
		this.bpelProcessMap = bpelProcessMap;
		this.environmentFactory = environmentFactory;
	}

	/**
	 * TODO USED This method is realize when the su is deployed
	 *
	 * @param serviceUnitName
	 *            the name of the service unit
	 * @param jbiSuRootPath
	 *            the su root path
	 * @param jbiDescriptor
	 *            the JBI descriptor
	 * @throws PEtALSCDKException
	 *             Impossible to realize this method
	 */
	public void onSUDeployed(final String serviceUnitName,
			final String jbiSuRootPath, final Jbi jbiDescriptor)
			throws PEtALSCDKException {
		try {
			this.logger
					.info("Perform BPEL-Processes deployment on Orchestra server");
			this.descriptor = jbiDescriptor;
			this.suRootPath = jbiSuRootPath;
			this.createBPELProcessEnvironmentsForServiceUnit(this.descriptor);

			// Deploy the BPEL process
			final List<Provides> providesList = jbiDescriptor.getServices()
					.getProvides();
			for (final Provides provides : providesList) {
				final BPELProcessEnvironment bpelProcessEnv = this.bpelProcessMap
						.get(provides.getServiceName()
								+ provides.getEndpointName());
				deployProcessToOrchestra(bpelProcessEnv);
				this.logger
						.info("The BPEL Process "
								+ bpelProcessEnv.bpelProcessName
								+ " has been succesfully deployed on Orchestra server and is available on the endpoint "
								+ provides.getServiceName()
								+ provides.getEndpointName());
			}

			this.logger
					.info("BPEL-Processes deployment on Orchestra server has been performed");
		} catch (Exception e) {
			e.printStackTrace();
			throw new PEtALSCDKException(e);
		}
	}

	/**
	 * This method is realize when the su is started
	 *
	 * @param serviceUnitName
	 *            the name of the service unit
	 * @throws OrchestraException
	 *             Impossible to realize this method
	 */
	public void onSUStarted(final String serviceUnitName)
			throws PEtALSCDKException {
	}

	/**
	 * this method is realize when the su is stopped
	 *
	 * @param serviceUnitName
	 *            the name of the service unit
	 * @throws OrchestraException
	 *             Impossible to realize this method
	 */
	public void onSUStopped(final String serviceUnitName)
			throws PEtALSCDKException {
	}

	/**
	 * this method is realize when the su is undeployed
	 *
	 * @param serviceUnitName
	 *            the name of the service unit
	 * @param jbiSuRootPath
	 *            the su root path
	 * @param jbiDescriptor
	 *            the JBI descriptor
	 * @throws PEtALSCDKException
	 *             Impossible to realize this method
	 */
	public void onSUUndeployed(final String serviceUnitName,
			final String jbiSuRootPath, final Jbi jbiDescriptor)
			throws PEtALSCDKException {
		this.logger
				.info("Perform BPEL-Processes undeployment on Orchestra server");

		// Undeploy the BPEL process
		final List<Provides> providesList = jbiDescriptor.getServices()
				.getProvides();
		for (final Provides provides : providesList) {
			final BPELProcessEnvironment bpelProcessEnv = this.bpelProcessMap
					.remove(provides.getServiceName()
							+ provides.getEndpointName());
			try {
				undeployProcessToOrchestra(bpelProcessEnv);
			} catch (final MessagingException e) {
				throw new PEtALSCDKException(e);
			}
			this.logger.info("The BPEL Process "
					+ bpelProcessEnv.bpelProcessName
					+ " has been succesfully undeployed on Orchestra server");
		}
		this.logger
				.info("BPEL-Processes undeployment on Orchestra server has been performed");
	}

	/**
	 * TODO USED create BPELProcess Environments for each Provide element in the
	 * SU. the BPELProcess Environments are kept in the
	 * BPELProcessEnvironmentsMap
	 *
	 * @param jbiDescriptor
	 *            the JBIDescriptor
	 * @throws PEtALSCDKException
	 *             impossible to generate urls
	 * @throws OrchestraException
	 */
	private void createBPELProcessEnvironmentsForServiceUnit(
			final Jbi jbiDescriptor) throws PEtALSCDKException {
		final List<Provides> providesList = jbiDescriptor.getServices()
				.getProvides();
		try {
			for (final Provides provides : providesList) {
				BPELProcessEnvironment bpelContext = null;
				ConfigurationExtensions props = new ConfigurationExtensions(
						provides.getAny());

				// get parameters from extension
				final String bpelProcessName = props
						.get(OrchestraSUListener.BPEL_PROPERTY);
				final String wsdlName = provides.getWsdl();

				final ArrayList<URL> externalWsdls = WSDLAnalyzer
						.getExternalWsdlUrls(this.suRootPath, WSDLAnalyzer
								.getResourceAsURL(this.suRootPath, wsdlName));

				bpelContext = this.createBPELProcessEnvironment(
						bpelProcessName, wsdlName, externalWsdls);

				// Store the BPEL process
				this.bpelProcessMap.put(provides.getServiceName()
						+ provides.getEndpointName(), bpelContext);
				this.logger
						.fine("A new BPELProcessEnvironment has been created for process "
								+ bpelProcessName);
			}
		} catch (final MalformedURLException e) {
			throw new PEtALSCDKException(e.getMessage());
		}
	}

	/**
	 * TODO used Create a BPEL process environment with the ServiceUnit
	 * parameters
	 *
	 * @param externalWsdls
	 *            list of all external WSDLs used by the process
	 * @throws MalformedURLException
	 *             impossible to generate URLs
	 * @return the BPEL context
	 */
	private BPELProcessEnvironment createBPELProcessEnvironment(
			final String bpelProcessName, final String wsdlName,
			final ArrayList<URL> externalWsdls) throws MalformedURLException {

		if (bpelProcessName == null || bpelProcessName.equals("")) {
			throw new IllegalArgumentException(
					"the BPEL process cannot be null in jbi descriptor");
		}
		if (wsdlName == null || wsdlName.equals("")) {
			throw new IllegalArgumentException(
					"the WSDL name cannot be null in jbi descriptor");
		}

		return new BPELProcessEnvironment(WSDLAnalyzer.getResourceAsURL(
				this.suRootPath, bpelProcessName), WSDLAnalyzer
				.getResourceAsURL(this.suRootPath, wsdlName),
				this.environmentFactory, externalWsdls);
	}

	/**
	 * Deploy the bpel process on the orchestra server
	 *
	 * @throws OrchestraException
	 *
	 * @throws OrchestraException
	 *             impossible to deploy the process
	 */
	public void deployProcessToOrchestra(BPELProcessEnvironment bpelProcessEnv)
			throws OrchestraException {
		if (bpelProcessEnv.bpelProcess == null) {
			try {
				this.logger.fine("Open the BPEL context");
				bpelProcessEnv.openContext();
				this.logger.fine("Deployment of the BPEL process "
						+ bpelProcessEnv.bpelProcessURL
						+ " on Orchestra server");
				Deployer bpelDeployer = new Deployer();

				final List<URL> wsdls = new ArrayList<URL>();

				for (final URL externalWsdl : bpelProcessEnv.externalsWsdls) {
					wsdls.add(externalWsdl);
				}
				if (bpelProcessEnv.wsdlURL != null) {
					wsdls.add(bpelProcessEnv.wsdlURL);
				}

				Deployment bpelDeployment = BpelImportParser.parseBpelFromUrl(
						bpelProcessEnv.bpelProcessURL, wsdls);

				bpelProcessEnv.bpelProcess = bpelDeployer.deploy(
						bpelDeployment, bpelProcessEnv.jbiEnvironment);

				if (bpelProcessEnv.bpelProcess == null) {
					throw new OrchestraException(
							"Process can not be deployed to Orchestra");
				}
			} catch (Exception e) {
				if (bpelProcessEnv.jbiEnvironment != null) {
					Transaction t = bpelProcessEnv.jbiEnvironment
							.getTransaction();
					if (t != null) {
						t.setRollbackOnly();
					}
				}
				throw new OrchestraException(e);
			} finally {
				if (bpelProcessEnv.jbiEnvironment != null) {
					bpelProcessEnv.jbiEnvironment.close();
				}
			}
			this.logger.fine("BPEL process " + bpelProcessEnv.bpelProcessURL
					+ " deployed on Orchestra server");

		}
	}

	/**
	 * Undeploy the bpel process from orchestra server
	 *
	 * @throws MessagingException
	 *
	 * @throws OrchestraException
	 *             error during process undeployment
	 */
	public void undeployProcessToOrchestra(BPELProcessEnvironment bpelProcessEnv)
			throws MessagingException, OrchestraException {
		try {
			if (bpelProcessEnv.bpelProcess != null) {
				Deployer bpelDeployer = new Deployer();
				bpelDeployer.undeploy(bpelProcessEnv.bpelProcess.getQName(),
						bpelProcessEnv.jbiEnvironment);
				bpelProcessEnv.closeContext();
				this.logger
						.info("BPEL process "
								+ bpelProcessEnv.bpelProcess.getQName()
								+ " undeployed");
			}
		} catch (final Exception e) {
			throw new MessagingException(e);
		}
	}

}
