/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.engine.orchestra.listener;

import java.util.Map;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;

import org.ow2.orchestra.exception.Fault;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.ExchangeUtil;
import org.ow2.petals.engine.orchestra.BPELProcessEnvironment;
import org.ow2.petals.engine.orchestra.OrchestraEngine;
import org.ow2.petals.engine.orchestra.OrchestraEngineException;
import org.ow2.petals.engine.orchestra.jbi.JBIReceiver;
import org.ow2.petals.engine.orchestra.util.Marshaller;

/**
 * This class implements the JBI listener for orchestra engine
 *
 * @author alouis
 * @author Nicolas SALATGE - eBMWebSourcing
 *
 */
public final class JBIListener extends AbstractJBIListener {

    private Map<String, BPELProcessEnvironment> bpelProcessMap;

    /**
     * The default constructor
     */
    public JBIListener() {

    }

    /**
     * Retrieve the BPELProcessEnvironmentMap for further access to bpel process
     * environments
     *
     * @see org.ow2.petals.component.common.listener.AbstractListener#init()
     */
    @Override
    protected void init() {
        this.bpelProcessMap = ((OrchestraEngine) this.getComponent())
        .getBPELProcessEnvironmentMap();
    }
    
    public class Process extends Thread {
        private final BPELProcessEnvironment bpelProcessEnv;

        private Exception exception;

        private final Exchange exchangeCopy;
        
        public Process(final BPELProcessEnvironment bpelProcessEnv, final Exchange exchangeCopy) {
            this.bpelProcessEnv = bpelProcessEnv;
            this.exchangeCopy = exchangeCopy;
        }

        @Override
        public void run() {
            try {
                JBIReceiver.onJBIMessageReceived(bpelProcessEnv, exchangeCopy);
            } catch (final Exception e) {
                if (exchangeCopy.isRobustInOnlyPattern()) {
                    this.exception = e;
                }
                getLogger().log(Level.SEVERE, e.getMessage(), e); 
            }
            getLogger().info(
                    "the " + bpelProcessEnv.bpelProcessName
                            + " BPEL process call is terminated for the Exchange "
                            + exchangeCopy.getExchangeId());
        }

        public Exception getMessageStatus() {
            return this.exception;
        }
    }
    
    /**
     * (non-Javadoc)
     *
     * @param exchange
     *            the exchange
     * @see org.ow2.petals.component.common.listener.AbstractJBIListener#onJBIMessage(org.ow2.petals.component.common.util.MessageExchangeWrapper)
     */
    @Override
    public void onJBIMessage(final Exchange exchange) {
        String endpointFullName = exchange.getEndpoint().getServiceName()
        + exchange.getEndpoint().getEndpointName();
        final BPELProcessEnvironment bpelProcessEnv = bpelProcessMap
        .get(endpointFullName);

        // retrieve the bpelProcess environment
        if (bpelProcessEnv == null) {
            String msg = "No process definition found for endpoint "
                + endpointFullName;
            getLogger().log(Level.SEVERE, msg);
            exchange.setError(new MessagingException(msg));
        } else {
            try {

                if ((exchange.isInOnlyPattern())
                        || (exchange.isRobustInOnlyPattern())) {

                    getLogger().info(
                            "Start a new instance or continue an existing instance of the process "
                            + bpelProcessEnv.bpelProcessName+ " for the Exchange "+exchange.getExchangeId());
                    getLogger().info("no response expected");

                    this.getLogger().finest(
                    "copy the exchange to use it in a separate thread");
                    final Exchange exchangeCopy = ExchangeUtil.copyExchange(
                            this.getComponent().getChannel()
                            .createExchangeFactory(
                                    exchange.getEndpoint()), exchange);

                    Process launchProcess = new Process(bpelProcessEnv, exchangeCopy);
                    launchProcess.run();
                    if (launchProcess.getMessageStatus() != null) {
                        this.setFaultOrError( exchange,
                                bpelProcessEnv, launchProcess.getMessageStatus());
                    }
                    getLogger()
                    .info(
                    "the process instance is executed in a new thread, the JBI exchange is terminated");

                } else {
                    getLogger().info(
                            "Start a new instance or continue an existing instance of the process "
                            + bpelProcessEnv.bpelProcessName + " for the Exchange "+exchange.getExchangeId());
                    getLogger().info(
                    "and wait for the process instance response");
                    JBIReceiver.onJBIMessageReceived(bpelProcessEnv, exchange);

                    getLogger().info(
                            "the " + bpelProcessEnv.bpelProcessName
                            + " BPEL process instance is terminated for the Exchange "+exchange.getExchangeId());
                }
            } catch (final OrchestraEngineException e) {
                getLogger()
                .log(
                        Level.SEVERE,
                        "OrchestraEngineException: Message can not be processed, set the Exchange status as Error",
                        e);
                exchange.setError(new MessagingException(e.getMessage()));
            } catch (final OrchestraException e) {
                getLogger()
                .log(
                        Level.SEVERE,
                        "OrchestraException: Message can not be processed",
                        e);
                this.setFaultOrError(exchange, bpelProcessEnv, e);
            } catch (final MessagingException e) {
                getLogger()
                .log(
                        Level.SEVERE,
                        "MessagingException: Message can not be processed, set the Exchange status as Error",
                        e);
                exchange.setError(e);
            } catch (PEtALSCDKException e) {
                getLogger()
                .log(
                        Level.SEVERE,
                        "PEtALSCDKException: Message can not be processed, set the Exchange status as Error",
                        e.getMessage());
                exchange.setError(new MessagingException(e));
            }
        }
    }

    private void setFaultOrError(final Exchange exchange,
            final BPELProcessEnvironment bpelProcessEnv, final Exception e) {
        if(e instanceof Fault) {
            try {
                getLogger()
                .log(
                        Level.SEVERE,
                        "OrchestraException: set the Exchange status as Fault",
                        e);
                javax.jbi.messaging.Fault fault = Marshaller.buildJBIFaultMessagesFromBPELFaultMessages(exchange, bpelProcessEnv, (Fault) e);
                exchange.setFault(fault);
            } catch (PEtALSCDKException e1) {
                getLogger()
                .log(
                        Level.SEVERE,
                        "OrchestraException: Cannot process the fault, set the Exchange status as Error",
                        e);
                exchange.setError(new MessagingException(e));
            } catch (MessagingException e1) {
                getLogger()
                .log(
                        Level.SEVERE,
                        "OrchestraException: Cannot process the fault, set the Exchange status as Error",
                        e);
                exchange.setError(e);
            }
        } else {
            getLogger()
            .log(
                    Level.SEVERE,
                    "OrchestraException: set the Exchange status as Error",
                    e);
            exchange.setError(new MessagingException(e.getMessage()));
        }
    }


}
