/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.engine.orchestra.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.security.auth.Subject;
import javax.wsdl.Fault;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap12.SOAP12Address;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.jbpm.pvm.internal.util.XmlUtil;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.var.BpelFaultMessage;
import org.ow2.orchestra.var.Message;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.util.SOAPFault;
import org.ow2.petals.component.framework.util.SOAPFault.BindingSoapVersionConstants;
import org.ow2.petals.component.framework.util.SOAPFault.SoapFaultCodeConstants;
import org.ow2.petals.engine.orchestra.BPELProcessEnvironment;
import org.ow2.petals.engine.orchestra.OrchestraEngineException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 *
 * @author nsalatge
 *
 */

public final class Marshaller {

	/**
	 * Mode
	 */
	public static enum ExchangeStyle {
		RPC, DOCUMENT
	};

	public MessageExchangeFactory exchangeFactory;

	/**
	 * Singleton
	 */
	private static Marshaller marshaller;

	/**
	 * Default constructor
	 */
	private Marshaller() {

	}

	/**
	 * Get instance of the marshaller
	 *
	 * @return the marshaller
	 */
	public static Marshaller getInstance() {
		if (Marshaller.marshaller == null) {
			Marshaller.marshaller = new Marshaller();
		}
		return Marshaller.marshaller;
	}

	public void init(MessageExchangeFactory exchangeFactory) {
		this.exchangeFactory = exchangeFactory;
	}

	/**
	 * Create a JBI message exchange
	 *
	 * @param itf
	 *            the QName
	 * @param operation
	 *            the operation name
	 * @param content
	 *            the content of the operation
	 * @param attachmentFiles
	 *            the list of attachments
	 * @param listener
	 *            the listener
	 * @param properties
	 *            the properties
	 * @param subject
	 *            the orchestra subject
	 * @return the exchange. May be null
	 * @throws MessagingException
	 * @throws PEtALSCDKException
	 */
	private Exchange createMsg(final QName itf, final Operation operation,
			final Document content, final List<File> attachmentFiles,
			final String properties, final Subject subject)
	throws MessagingException, PEtALSCDKException {

		MEPConstants mep = MEPConstants.IN_ONLY_PATTERN;

		if (operation.getOutput() != null) {
			mep = MEPConstants.IN_OUT_PATTERN;
		}
		final Exchange exchange = createExchange(mep);
		exchange.setInterfaceName(itf);

		if (exchange != null) {
			exchange.setOperation(QName.valueOf(operation.getName()));
			// setting content
			exchange.setInMessageContent(content);

			if (subject != null) {
				exchange.getInMessage().setSecuritySubject(subject);

			}

			// TODO: Copy attachment and other properties
		}
		return exchange;
	}

	/**
	 * Build JBI message from BPEL message
	 *
	 * @param wsdl
	 *            the WSDL
	 * @param message
	 *            the BPEL Message
	 * @param portTypeQName
	 *            the portTypeQName
	 * @param operation
	 *            the operation
	 * @param wsdlPart
	 *            the orchestra WSDL collection
	 * @param itf
	 *            the QName
	 * @param subject
	 *            the orchestra subject
	 * @return the jbi message exchange
	 * @throws OrchestraEngineException
	 *             impossible to build JBI message
	 * @throws WSDLException
	 */
	public Exchange createJBIMessageExchangeFromBPELMessage(
			final WsdlsInfos wsdl, final QName portTypeQName,
			final Message message, final Operation operation,
			final Collection<Part> wsdlPart, final QName itf,
			final Subject subject) throws OrchestraEngineException, WSDLException {
		Exchange exchange = null;
		try {
			final Document document = this.buildJBIMessageFromBPELMessage(wsdl,
					portTypeQName, message, operation, wsdlPart, itf
					.getNamespaceURI());

			exchange = this.createMsg(itf, operation, document, null, null,
					subject);

		} catch (final MessagingException e) {
			throw new OrchestraEngineException(
					"Exception while building a JBI Message", e);
		} catch (PEtALSCDKException e) {
			throw new OrchestraEngineException(
					"Exception while building a JBI Message", e);
		}
		return exchange;
	}

	/**
	 * Build JBI Document from BPEL message
	 *
	 * @param wsdl
	 *            the WSDL
	 * @param portTypeQName
	 *            the portTypeQName
	 * @param message
	 *            the BPEL Message
	 * @param operation
	 *            the operation
	 * @param wsdlPart
	 *            the WSDL collection
	 * @param operationNamespace
	 *            the operationNamespace
	 * @return the JBI document
	 * @throws OrchestraEngineException
	 *             impossible to build JBI document
	 * @throws MarshallingException
	 */
	public Document buildJBIMessageFromBPELMessage(final WsdlsInfos wsdl,
			final QName portTypeQName, final Message message,
			final Operation operation, final Collection<Part> wsdlPart,
			final String operationNamespace) throws MarshallingException {
		Document document = null;

		WSDLAnalyzer analyzer;

		analyzer = new WSDLAnalyzer(wsdl);

		final ExchangeStyle style = analyzer.getStyleForOperation(
				portTypeQName, operation.getName());

		if (style == null) {
			throw new MarshallingException(
			"the message style (rpc or document) can not be found");
		}

		document = BpelUtil.getNewDocument();

		final List<Element> parts = this.getListOfParts(message, document);

		if (style == ExchangeStyle.RPC) {
			final Element operationElmt = document.createElementNS(
					operationNamespace, operation.getName());
			operationElmt.setPrefix("s");
			for (final Part part : wsdlPart) {
				Element partElmt = findCorrespondingPart(part, parts);

				Element elmt = null;
				elmt = this.importElementInJbiMessage(portTypeQName, wsdlPart,
						document, partElmt, style);
				if (elmt == null) {
					throw new MarshallingException("the part message "
							+ partElmt.getNodeName()
							+ " can not be found in the operation description"
							+ operation.getName());
				}
				operationElmt.appendChild(elmt);
			}
			document.appendChild(operationElmt);
		} else {
			final List<Element> elmts = new ArrayList<Element>();
			for (final Element part : parts) {
				Element elmt = null;
				elmt = this.importElementInJbiMessage(portTypeQName, wsdlPart,
						document, part, style);
				if (elmt == null) {
					throw new MarshallingException("the part message "
							+ part.getNodeName()
							+ " can not be found in the operation description"
							+ operation.getName());
				}

				elmts.add(elmt);
			}

			if (elmts.size() > 1) {
				throw new MarshallingException(
				"the document has more than one root element");
			}

			if (elmts.size() == 1) {
				document.appendChild(elmts.get(0));
			}
		}
		return document;
	}

	private Element findCorrespondingPart(Part part, List<Element> parts) {
		Element elmt = null;
		for (final Element partElmt : parts) {
			if (part.getName().equals(partElmt.getNodeName())) {
				elmt = partElmt;
				break;
			}
		}
		return elmt;
	}

	private Element importElementInJbiMessage(final QName portTypeQName,
			final Collection<Part> wsdlPart, final Document document,
			final Element part, final ExchangeStyle style) {
		Element elmt = null;
		for (final Part p : wsdlPart) {
			if (p.getName().equals(part.getNodeName())) {
				if (p.getElementName() != null) {
					elmt = document.createElementNS(p.getElementName()
							.getNamespaceURI(), p.getElementName()
							.getLocalPart());
				} else {
					if (style == ExchangeStyle.RPC) {
						elmt = document.createElement(p.getName());
					} else {
						elmt = document.createElementNS(portTypeQName
								.getNamespaceURI(), p.getName());
					}
				}
				if (part.getFirstChild() != null
						&& part.getFirstChild().getNodeType() == Node.TEXT_NODE
						&& part.getChildNodes().getLength() == 1) {
					elmt.setTextContent(part.getTextContent());
				} else {
					for (int i = part.getChildNodes().getLength() - 1; i >= 0; i--) {
						final Node newNode = elmt.getOwnerDocument().adoptNode(
								part.getChildNodes().item(i).cloneNode(true));
						elmt.appendChild(newNode);
					}
				}

				break;
			}
		}
		return elmt;
	}

	private List<Element> getListOfParts(final Message message,
			final Document document) {
		final List<Element> parts = new ArrayList<Element>();
		final Iterator<Entry<String, Element>> it = message.getParts()
		.entrySet().iterator();
		while (it.hasNext()) {
			final Entry<String, Element> entry = it.next();
			final String partName = entry.getKey();
			final Element partElement = entry.getValue();

			final Element partEl = document.createElement(partName);

			if (partElement.getFirstChild() != null
					&& partElement.getFirstChild().getNodeType() == Node.TEXT_NODE
					&& partElement.getChildNodes().getLength() == 1) {
				partEl.setTextContent(partElement.getTextContent());

			} else {
				for (int i = partElement.getChildNodes().getLength() - 1; i >= 0; i--) {
					final Node newNode = partEl.getOwnerDocument()
					.adoptNode(
							partElement.getChildNodes().item(i)
							.cloneNode(true));
					partEl.appendChild(newNode);
				}

			}
			parts.add(partEl);
		}
		return parts;
	}

	/**
	 * Build BPEL message from JBI message exchange
	 *
	 * @param parts
	 *            the bpel parts element
	 * @param jbiMessage
	 *            the jbi message
	 * @return the BPEL message
	 * @throws MarshallingException
	 *             problem while converting the message
	 */
	public Message buildBPELMessageFromJBIMessage(final WsdlsInfos wsdl,
			final QName portTypeQName, final Operation operation,
			final Collection<Part> parts, final Document jbiMessage)
	throws MarshallingException {
		Message message = null;

		final HashMap<String, Element> messageParts = new HashMap<String, Element>();

		for (final Part part : parts) {

			Element el = null;

			WSDLAnalyzer analyzer;

			analyzer = new WSDLAnalyzer(wsdl);

			final ExchangeStyle style = analyzer.getStyleForOperation(
					portTypeQName, operation.getName());

			if (style == null) {
				throw new MarshallingException(
				"the message style (rpc or document) can not be found");
			}

			if (part.getElementName() != null) {
				el = this.getPartOrElement(jbiMessage.getDocumentElement(),
						part.getElementName().getLocalPart(), style);
			} else {
				el = this.getPartOrElement(jbiMessage.getDocumentElement(),
						part.getName(), style);
			}

			if (el == null) {

				if(style == ExchangeStyle.RPC) {
					if (part.getElementName() != null) {
						el = this.getPartOrElement(jbiMessage.getDocumentElement(),
								part.getElementName().getLocalPart(), ExchangeStyle.DOCUMENT);
					} else {
						el = this.getPartOrElement(jbiMessage.getDocumentElement(),
								part.getName(), ExchangeStyle.DOCUMENT);
					}
				} else {
					if (part.getElementName() != null) {
						el = this.getPartOrElement(jbiMessage.getDocumentElement(),
								part.getElementName().getLocalPart(), ExchangeStyle.RPC);
					} else {
						el = this.getPartOrElement(jbiMessage.getDocumentElement(),
								part.getName(), ExchangeStyle.RPC);
					}
				}


				if(el == null) {

					String partOrElmt = null;
					if (part.getElementName() != null) {
						partOrElmt = part.getElementName().getLocalPart();
					}
					if (partOrElmt == null) {
						partOrElmt = part.getName();
					}

					throw new MarshallingException("the part or element named "
							+ partOrElmt + " is not found in the request");
				}
			}

			final Element newElement = this.copyElementWithNs(part, el);

			if (part.getTypeName() != null
					&& BpelUtil.isXmlSchemaNs(part.getTypeName()
							.getNamespaceURI())) {
				// it is a schema type
				String text = null;
				final Node child = newElement.getFirstChild();

				if (child != null) {
					if (child.getNodeType() == Node.TEXT_NODE) {
						text = ((Text) child).getData();
					} else if (child.getNodeType() == Node.ELEMENT_NODE) {
						text = ((Element) child).getTextContent();
					}
				} else {
					text = newElement.getTextContent();
				}
				final Element element = BpelXmlUtil
				.createElementWithContent(text);
				messageParts.put(part.getName(), element);
			} else {
				messageParts.put(part.getName(), newElement);
			}

		}

		message = new Message(messageParts);

		return message;
	}

	/**
	 * Build BPEL message from JBI message exchange
	 *
	 * @param parts
	 *            the bpel parts element
	 * @param jbiMessage
	 *            the jbi message
	 * @return the BPEL message
	 * @throws MarshallingException
	 *             can not convert the message
	 */
	public List<BpelFaultMessage> buildBPELFaultMessagesFromJBIFaultMessages(
			final WsdlsInfos wsdl, final QName portTypeQName,
			final Operation operation, final List<Element> soapFaults)
			throws MarshallingException {
		List<BpelFaultMessage> messages = null;

		final HashMap<String, Element> messageParts = new HashMap<String, Element>();
		@SuppressWarnings("unchecked")
		final Map<String, Fault> mapFaults = ((Map<String, Fault>) operation
				.getFaults());
		if (mapFaults != null) {

			WSDLAnalyzer analyzer;
			analyzer = new WSDLAnalyzer(wsdl);
			final ExchangeStyle style = analyzer.getStyleForOperation(
					portTypeQName, operation.getName());

			if (style == null) {
				throw new MarshallingException(
				"message style (rpc or document) can not be found");
			}

			Collection<Fault> faults = mapFaults.values();
			for (Fault fault : faults) {
				@SuppressWarnings("unchecked")
				final Map<String, Part> faultParts = fault.getMessage()
				.getParts();
				for (final Part part : faultParts.values()) {
					Element el = null;

					for (Element soapFault : soapFaults) {
						if (part.getElementName() != null) {
							el = this.getPartOrElement(soapFault, part
									.getElementName().getLocalPart(), style);
						} else {
							el = this.getPartOrElement(soapFault, part
									.getName(), style);
						}

						if (el != null) {

							final Element newElement = el;

							if (part.getTypeName() != null
									&& BpelUtil.isXmlSchemaNs(part
											.getTypeName().getNamespaceURI())) {
								// it is a schema type
								String text = null;
								final Node child = newElement.getFirstChild();

								if (child != null) {
									if (child.getNodeType() == Node.TEXT_NODE) {
										text = ((Text) child).getData();
									} else if (child.getNodeType() == Node.ELEMENT_NODE) {
										text = ((Element) child)
										.getTextContent();
									}
								} else {
									text = newElement.getTextContent();
								}
								final Element element = BpelXmlUtil
								.createElementWithContent(text);
								messageParts.put(part.getName(), element);
							} else {
								messageParts.put(part.getName(), newElement);
							}

							if (messages == null) {
								messages = new ArrayList<BpelFaultMessage>();
							}

							messages.add(new BpelFaultMessage(new QName(
									portTypeQName.getNamespaceURI(), fault
									.getName()), messageParts));
						}
					}

				}
			}
		}

		return messages;
	}
	
	/**
	 *     
     * Build JBI Fault message from BPEL Fault message exchange

	 * @param exchange the exchange
	 * @param bpelProcessEnv the process
	 * @param e the bpel exception
	 * @return the jbi fault
	 * @throws MessagingException
	 * @throws PEtALSCDKException
	 */
    @SuppressWarnings("unchecked")
    public static javax.jbi.messaging.Fault buildJBIFaultMessagesFromBPELFaultMessages(final Exchange exchange,
            final BPELProcessEnvironment bpelProcessEnv, final org.ow2.orchestra.exception.Fault e)
            throws MessagingException, PEtALSCDKException {
        // get list of detail
        List<Element> details = new ArrayList<Element>();
        List<Element> bpelfaults = null;
        for (BpelFaultMessage fault : e.getFaults()) {
            for (Element element : fault.getParts().values()) {
                bpelfaults = XmlUtil.elements(element);
                if((bpelfaults != null)&&(bpelfaults.size()>0)) {
                    details.add(bpelfaults.get(0));
                }
            }
        }
        
        // find soap binding and create source
        Source faultContent = null;
        Service serv = bpelProcessEnv.wsdlDef.getService(exchange.getService());
        Port endpoint = null;
        for(Port ep: ((Map<QName, Port>)serv.getPorts()).values()) {
            for (ExtensibilityElement extensibilityElement : (List<ExtensibilityElement>)ep.getExtensibilityElements()) {
                if(extensibilityElement.getElementType().getLocalPart().equals("address")) {
                    if(extensibilityElement instanceof SOAPAddress) {
                        if(((SOAPAddress)extensibilityElement).getLocationURI().equals(exchange.getEndpointName())) {
                            endpoint = ep;
                            break;
                        }
                    } else if(extensibilityElement instanceof SOAP12Address) {
                        if(((SOAP12Address)extensibilityElement).getLocationURI().equals(exchange.getEndpointName())) {
                            endpoint = ep;
                            break;
                        }
                    } 
                }
            }
        }
        
        if(endpoint == null) {
            throw new PEtALSCDKException("Error: impossible to find soap binding");
        }
        
        List<ExtensibilityElement> elmt = endpoint.getBinding().getExtensibilityElements();
        for (ExtensibilityElement extensibilityElement : elmt) {
            if(extensibilityElement.getElementType().getLocalPart().equals("binding")) {
                if(extensibilityElement.getElementType().getNamespaceURI().equals(BindingSoapVersionConstants.SOAP11_NS.toString())) {
                    faultContent = SOAPFault.createSOAP11Fault(SoapFaultCodeConstants.SERVER, "bpel process fail: " + exchange.getService(), details);
                    break;
                } else if(extensibilityElement.getElementType().getNamespaceURI().equals(BindingSoapVersionConstants.SOAP12_NS.toString())) {
                    faultContent = SOAPFault.createSOAP12Fault(SoapFaultCodeConstants.SERVER, "bpel process fail: " + exchange.getService(), details);
                    break;
                } 
            }
        }
        
        if(faultContent == null) {
            throw new PEtALSCDKException("Error: impossible to find soap binding");
        }
        
        javax.jbi.messaging.Fault fault = exchange.createFault();
        fault.setContent(faultContent);
        return fault;
    }

	/**
	 * Copy of elements with their name space
	 *
	 * @param part
	 * @param el
	 * @return
	 * @throws MarshallingException
	 */
	private Element copyElementWithNs(final Part part, final Element el)
	throws MarshallingException {
		Element newElement = null;

		if (part.getTypeName() != null) {
			newElement = BpelXmlUtil.createElement(part.getTypeName()
					.getNamespaceURI(), part.getName());
		} else if (part.getElementName() != null) {
			newElement = BpelXmlUtil.createElement(part.getElementName()
					.getNamespaceURI(), part.getName());
		}

		if (newElement == null) {
			throw new MarshallingException(
					"the type or the element associate to the following part can not be found : "
					+ part.getName());
		}

		for (int i = 0; i < el.getChildNodes().getLength(); i++) {
			final Node n = newElement.getOwnerDocument().adoptNode(
					el.getChildNodes().item(i).cloneNode(true));

			newElement.appendChild(n);
		}

		return newElement;
	}

	/**
	 * Find the part or the element of a father element from the father element
	 *
	 * @param father
	 *            the father element
	 * @param partOrElmtName
	 *            the part / element name
	 * @return return null
	 */
	// TODO correct the number of 'return' (put only one instead of 3)
	private Element getPartOrElement(final Element father,
			final String partOrElmtName, final ExchangeStyle style) {

		if ((style != ExchangeStyle.RPC)
				&& (father.getLocalName().equals(partOrElmtName))) {
			return father;
		}

		for (int i = 0; i < father.getChildNodes().getLength(); i++) {
			if (father.getChildNodes().item(i).getLocalName() != null
					&& father.getChildNodes().item(i).getLocalName().equals(
							partOrElmtName)) {
				return (Element) father.getChildNodes().item(i);
			}
		}
		return null;
	}

	/**
	 * FIXME use some CDK util methods to create the exchange Create an empty
	 * Exchange
	 *
	 * @param mep
	 *            exchange pattern
	 * @return
	 * @throws MessagingException
	 *             JBI factory can not create the MsgExchange
	 */
	public Exchange createExchange(MEPConstants mep) throws MessagingException {
		if (mep == null) {
			throw new IllegalArgumentException(
			"'mep' parameter must be non null");
		}
		MessageExchange exchange = this.exchangeFactory.createExchange(mep
				.value());
		return new ExchangeImpl(exchange);
	}
}
