/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.engine.orchestra.util;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.ow2.orchestra.exception.Error;
import org.ow2.orchestra.exception.Fault;
import org.ow2.orchestra.var.BpelFaultMessage;
import org.ow2.orchestra.var.Message;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.util.SOAPFault;
import org.ow2.petals.engine.orchestra.OrchestraEngineException;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;

/**
 * @author alouis
 * @author nsalatge
 *
 */

public class Sender {

    private Logger logger;

    private DeliveryChannel channel;

    private ComponentContext context;

    /**
     * Send style
     *
     */
    public static enum SendStyle {
        ASYNCHRONOUS_SEND, SYNCHRONOUS_SEND, SYNCHRONOUS_SEND_WITH_TIMEOUT
    };

    /**
     * Default constructor
     *
     */
    public Sender(final ComponentContext context,
            final DeliveryChannel channel, final Logger logger) {

        this.logger = logger;
        this.channel = channel;
        this.context = context;
    }

    /**
     * FIXME some orchestra Error and Fault exception are thrown FIXME use some
     * CDK utils to send Exchanges Call a jbi service
     *
     * @param partner
     *            the partner to invoke. May be null
     * @param msgIn
     *            the exchange
     * @param syncTime
     *            timeout for syncSend
     * @return the exchange response
     * @throws MessagingException
     *             problem while accessing MsgExch information or while sending
     *             the MsgExc
     * @throws OrchestraEngineException no endpoint found to send the message
     */
    public final Message send(final Element partner, final Exchange msgIn,
            final Long syncTime, WsdlsInfos infos, QName portTypeQName,
            Operation operation) {
        Message messageReturned = null;
        try {
            SendStyle sendStyle;
            sendStyle = SendStyle.ASYNCHRONOUS_SEND;
            if (operation.getOutput() != null) {
                sendStyle = SendStyle.SYNCHRONOUS_SEND;
            }
            
            this.logger.finest("Send a JBI message");
            String endpointName = null;
            QName service = null;

            this.logger.finest("message mep : " + msgIn.getPattern());
            if (partner != null) {
                endpointName = this.getEndPointName(partner);
                service = this.getService(partner);

                if (service != null) {
                    this.logger.info("Set the service to invoke: " + service);
                    msgIn.setService(service);
                    if (endpointName != null) {
                        ServiceEndpoint endpoint = this.resolveEndPoint(service,
                                endpointName);

                        this.logger.info("Set the endpoint to invoke: "
                                + endpoint.getEndpointName());
                        msgIn.setEndpoint(endpoint);
                    }
                }
            } else {
                this.logger
                .finest("partner is null: research the service by its interface");
            }

            this.logger.finest("Satus: " + msgIn.getStatus());
            this.logger.finest("Consumer Role? " + msgIn.isConsumerRole());

            if ((sendStyle == SendStyle.ASYNCHRONOUS_SEND)
                    && (msgIn.isInOnlyPattern())) {
                channel.send(((ExchangeImpl) msgIn).getMessageExchange());
            } else if (sendStyle == SendStyle.SYNCHRONOUS_SEND) {
                if (syncTime != null) {
                    channel.sendSync(((ExchangeImpl) msgIn).getMessageExchange(),
                            syncTime);
                } else {
                    channel.sendSync(((ExchangeImpl) msgIn).getMessageExchange());
                }

                if ((msgIn.getFault() != null) || (msgIn.getError() != null)
                        || (msgIn.isErrorStatus())) {
                    if (msgIn.getError() != null) {
                        throw new Error(msgIn.getError());
                    }

                    if (msgIn.getFault() != null) {
                        try {
                            SOAPFault soapfault = new SOAPFault(msgIn.getFault());

                            List<BpelFaultMessage> bpelfaults = Marshaller
                            .getInstance()
                            .buildBPELFaultMessagesFromJBIFaultMessages(
                                    infos, portTypeQName, operation,
                                    soapfault.getSoapFaults());

                            if (bpelfaults == null) {
                                throw new Error(
                                        "these received fault messages are not defined in WSDL document: "
                                        + soapfault.getSoapFaults());
                            }

                            throw new Fault(bpelfaults);
                        } catch (PEtALSCDKException e) {
                            throw new Error(
                            "Can not convert fault into XML document");
                        }
                    }
                    throw new Error("Error from the provider " + service
                            + endpointName);
                }

                this.logger.log(Level.INFO, "Receive response from: " + service
                        + endpointName);
                if (operation.getOutput() != null) {
                    @SuppressWarnings("unchecked")
                    final Map<String, Part> outputParts = operation.getOutput()
                    .getMessage().getParts();
                    messageReturned = Marshaller.getInstance()
                    .buildBPELMessageFromJBIMessage(infos, portTypeQName,
                            operation, outputParts.values(),
                            msgIn.getOutMessageContentAsDocument());

                    this.logger.finest("bpel message returned: " + messageReturned);
                }

                if (!msgIn.getPattern().equals(MEPConstants.IN_ONLY_PATTERN)) {
                    this.logger.finest("Set status done");
                    msgIn.setDoneStatus();
                    this.logger.finest("Send status done");
                    channel.send(((ExchangeImpl) msgIn).getMessageExchange());
                }

            } else {
                // FIXME should not happen : the exchange mep has been created in
                // correlation with the senStyle parameter in a previous method
                throw new Error(
                        "this kind of message ("
                        + sendStyle
                        + ") is unknown or is incompatible with the mep of message: "
                        + msgIn.getPattern());
            }

            this.logger.finest("JBI message sent");
        } catch (OrchestraEngineException e) {
            this.logger.severe(e.getMessage());
            throw new Error(e);
        } catch (MessagingException e) {
            this.logger.severe(e.getMessage());
            throw new Error(e);
        }
        return messageReturned;
    }

    /**
     * Extract the service from a partner reference information item.
     *
     * @param ii
     *            Information Item
     * @return the service QName
     */
    public final QName getService(final Element ii) {

        logger.log(Level.FINEST, "Entering in getService method");

        final org.w3c.dom.Element serviceRefElem = (org.w3c.dom.Element) ii;

        final org.w3c.dom.Element endpointRef = (org.w3c.dom.Element) serviceRefElem
        .getElementsByTagName("EndpointReference").item(0);

        final org.w3c.dom.Element serviceElmt = (org.w3c.dom.Element) endpointRef
        .getElementsByTagName("ServiceName").item(0);

        String serviceName = null;
        String prefixName = null;

        if (serviceElmt.getTextContent().indexOf(":") > -1) {
            serviceName = serviceElmt.getTextContent().substring(
                    serviceElmt.getTextContent().indexOf(":") + 1);
            prefixName = serviceElmt.getTextContent().substring(0,
                    serviceElmt.getTextContent().indexOf(":"));
        } else {
            serviceName = serviceElmt.getTextContent();
        }

        String serviceNamespace = serviceElmt.lookupNamespaceURI(prefixName);
        if (serviceNamespace == null) {
            String prefixTmp = null;
            final NamedNodeMap list = serviceElmt.getAttributes();
            for (int i = 0; i < list.getLength(); i++) {

                if (list.item(i).getNodeName().indexOf(":") > -1) {
                    prefixTmp = list.item(i).getNodeName().substring(
                            list.item(i).getNodeName().indexOf(":") + 1);
                } else {
                    prefixTmp = list.item(i).getNodeName();
                }

                if ((prefixName != null) && (prefixName.equals(prefixTmp))) {
                    serviceNamespace = list.item(i).getNodeValue();
                }
            }
        }

        // TODO: Normally, just this line is sufficient
        // QName service = new QName(serviceElmt.lookupNamespaceURI(prefixName),
        // serviceName);

        final QName service = new QName(serviceNamespace, serviceName);

        logger.log(Level.FINEST, "service name: " + service);
        logger.log(Level.FINEST, "Exit of the getService method");
        return service;
    }

    /**
     * Extract the endPoint reference from a service reference information item.
     *
     * @param ii
     *            Information Item
     * @return the end Point URL
     */
    public final String getEndPointName(final Element ii) {
        String endpointName = null;
        logger.log(Level.FINEST, "Entering in getEndPointName method");

        final org.w3c.dom.Element serviceRefElem = (org.w3c.dom.Element) ii;

        org.w3c.dom.Element endpointRef = null;

        if (serviceRefElem.getElementsByTagName("EndpointReference")
                .getLength() > 0) {
            endpointRef = (org.w3c.dom.Element) serviceRefElem
            .getElementsByTagName("EndpointReference").item(0);
        }

        if (endpointRef != null) {
            org.w3c.dom.Element endpoint = null;
            if (endpointRef.getElementsByTagName("Address").getLength() > 0) {
                endpoint = (org.w3c.dom.Element) endpointRef
                .getElementsByTagName("Address").item(0);
            }

            if (endpoint != null) {
                endpointName = endpoint.getTextContent();
            }
        }
        logger.log(Level.FINEST, "endpoint name found: " + endpointName);
        logger.log(Level.FINEST, "Exit of the getEndPointName method");

        return endpointName;
    }

    /**
     * Find the endpoint in the JBI env
     *
     * @param service
     * @param endpointName
     * @return
     * @throws OrchestraEngineException
     *             endpoint can not be found in the JBI context
     */
    public final ServiceEndpoint resolveEndPoint(final QName service,
            final String endpointName) throws OrchestraEngineException {

        logger.log(Level.FINEST, "find the endPoint in the JBI environment");

        logger.log(Level.FINEST, "service name: " + service);
        logger.log(Level.FINEST, "endpoint name: " + endpointName);

        // Search by endpoint
        ServiceEndpoint endpoint = context.getEndpoint(service, endpointName);

        if (endpoint == null) {
            throw new OrchestraEngineException("endpoint (" + service + "-"
                    + endpointName + ") not found in the JBI environment");
        }

        logger.log(Level.FINEST, "endpoint service found: " + endpoint);
        return endpoint;
    }

}
